/*
 *  SequentieelFile.java  -  Joris van Rantwijk
 *  in271 Inleiding databeheer
 *
 *  Bestandsorganisatie: Sequentieel file
 */

import java.io.RandomAccessFile;
import java.io.IOException;

/*
 *  We implementeren een sequentiele bestandsorganisatie.
 *  Er wordt gebruik gemaakt van 1 file, waarin alle records opgesomd worden
 *  in een willekeurige volgorde. Zoeken gebeurt door alle records af te
 *  lopen. Bij het wissen van een record wordt de sleutel op een ongeldige
 *  waarde (-1) gezet. Later kan een nieuw record over het ongeldige
 *  record geschreven worden.
 */

//===========================================================================
// Class SequentieelFile
// Implementatie van een sequentiele bestandsorganisatie
//===========================================================================
public class SequentieelFile extends BestandsOrganisatie
{
	// De file waarin de records worden opgeslagen.
	RandomAccessFile f;

	// Het aantal records in de file (niet allemaal geldig).
	int numRec;


	// Construeer een SequentieelFile.
	// Gebruik de gegeven filenaam, en het gegeven recordtype.
	//-----------------------------------------------------------------------
	public SequentieelFile(String filenaam, Class recordType)
		throws IOException
	{
		this.recordType = recordType;
		this.recordLen = Record.newRecord(recordType).getRecordlen();

		f = new RandomAccessFile(filenaam, "rw");
		numRec = (int) f.length() / recordLen;
	}


	// Sluit de file.
	//-----------------------------------------------------------------------
	public void close() throws IOException
	{
		f.close();
	}


	// Voeg het record r toe.
	// Probeer eerst een ongeldig record te hergebruiken, 
	// anders wordt r toegevoegd aan het eind van de file.
	//-----------------------------------------------------------------------
	public void insert(Record r) throws Exception, IOException
	{
		Record t;
		int i, k;

		// Construeer een nieuw object van het juiste recordtype.
		t = Record.newRecord(recordType);

		// Loop door het bestand en zoek naar een ongeldig record.
		// Kijk tegelijkertijd of een record met dezelfde sleutel bestaat.
		k = -1;
		for (i = 0; i < numRec; i++) {
			f.seek(i * recordLen);
			t.read(f);
			diskaccess++;
			if (t.key == r.key) {
				// Het record bestaat al, geef een exceptie.
				throw new Exception("Record bestaat al");
			}
			// Bewaar het nummer van een ongeldig record in k.
			if (k == -1 && t.key == -1) k = i;
		}

		// Als er geen ongeldig record is gevonden, voeg dan een record
		// toe aan het eind van de file.
		if (k == -1) k = numRec++;

		// Schrijf het record naar de file.
		f.seek(k * recordLen);
		r.write(f);
		diskaccess++;
	}


	// Vervang een bestaaand record door r.
	//-----------------------------------------------------------------------
	public void update(Record r) throws Exception, IOException
	{
		Record t;
		int i;

		// Construeer een nieuw object van het juiste recordtype.
		t = Record.newRecord(recordType);

		// Loop door het bestand en zoek naar een record met dezelfde key.
		for (i = 0; i < numRec; i++) {
			f.seek(i * recordLen);
			t.read(f);
			diskaccess++;
			if (t.key == r.key)
				break;
		}

		// Geef een exceptie als het record niet is gevonden.
		if (i == numRec) throw new Exception("Record niet gevonden");

		// Vervang het record door r.
		f.seek(i * recordLen);
		r.write(f);
		diskaccess++;
	}


	// Zoek het record met de sleutel key.
	//-----------------------------------------------------------------------
	public Record retrieve(int key) throws IOException
	{
		Record t;
		int i;

		// Construeer een nieuw object van het juiste recordtype.
		t = Record.newRecord(recordType);

		// Loop door het bestand en zoek een record met de juiste sleutel.
		for (i = 0; i < numRec; i++) {
			f.seek(i * recordLen);
			t.read(f);
			diskaccess++;
			if (t.key == key) break;
		}

		// Geef het gevonden record terug, of null als er geen
		// record is gevonden.
		if (i < numRec)
			return t;
		else
			return null;
	}

	
	// Zoek het record met de sleutel key en schrijf er een ongeldig
	// record overheen.
	//-----------------------------------------------------------------------
	public void delete(int key) throws Exception, IOException
	{
		Record t;
		int i;

		// Construeer een nieuw object van het juiste recordtype.
		t = Record.newRecord(recordType);

		// Loop door het bestand en zoek een record met de juiste sleutel.
		for (i = 0; i < numRec; i++) {
			f.seek(i * recordLen);
			t.read(f);
			diskaccess++;
			if (t.key == key) break;
		}

		// Geef een exceptie als het record niet is gevonden.
		if (i == numRec) throw new Exception("Record niet gevonden");

		// Vervang het record door een record met een ongeldige sleutel.
		t.key = -1;
		f.seek(i * recordLen);
		t.write(f);
		diskaccess++;
	}

}

//===========================================================================
// End
//===========================================================================
